"""Monte Carlo Particle Ray Tracing compiler, Volume 4"""
import mccode_antlr.grammar
__author__ = "Gregory Tucker"
__affiliation__ = "European Spallation Source ERIC"

from .version import version
__version__ = version()

__all__ = ["__author__", "__affiliation__", "__version__", "version"]
